/*-
 * slip.c - lots of blits
 *
 * Copyright (c) 1992 by Scott Draves (spot@cs.cmu.edu)
 *
 * See xlock.c for copying information.
 * 01-Dec-95: Patched for VMS <joukj@alpha.chem.uva.nl>.
 * 27-Dec-97: Patched for GGI <jchrr@hrz.uni-bielefeld.de>.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/time.h>
#ifdef __linux__
#include <linux/vt.h>
#endif
#include <string.h>
#include <ggi/libggi.h>
#include "ggi_screensaver.h"

#include <structs.h>

/* register screensavers */
void reg_savers();
void done_savers();
void ge_slip(ggi_visual_t visual);

/*
 * This structure contains all screensavers in this module:
 */
ggi_saver_list saver_list[] = {
	{ ge_slip, "Show slipping blits"}
};

/*
 * register all screen-savers in the main module
 */
void reg_savers()
{
	int i;
	for (i=0; i<(sizeof(saver_list)/sizeof(ggi_saver_list)); i++) {
		install_saver(saver_list[i].func, saver_list[i].name);
	}
}

/*
 * free resources allocated by the screen-savers
 */
void done_savers()
{
}  

/*
 * From here it's basically the old slip.c, with the X-calls replaced with
 * ggi-calls.
 */

#define LRAND lrand48

typedef struct {
	int         width, height;
	int         nblits_remaining;
	int         blit_width, blit_height;
	int         mode;
	int         first_time;
} slipstruct;

static slipstruct *slips = NULL;

static short
halfrandom(int mv)
{
	static short lasthalf = 0;
	unsigned long r;

	if (lasthalf) {
		r = lasthalf;
		lasthalf = 0;
	} else {
		r = LRAND();
		lasthalf = r >> 16;
	}
	return r % mv;
}

static int
erandom(int mv)
{
	static int  stage = 0;
	static unsigned long r;
	int         res;

	if (0 == stage) {
		r = LRAND();
		stage = 7;
	}
	res = r & 0xf;
	r = r >> 4;
	stage--;
	if (res & 8)
		return res & mv;
	else
		return -(res & mv);
}

static void
prepare_screen(ggi_visual_t visual, int xsize, int ysize, slipstruct * s)
{
	ggi_color map[256];

	int         c, i, n, w;
	int         not_solid;
	
	not_solid = halfrandom(10);
	w = s->width / 20;

	if (s->first_time || !halfrandom(5)) {
		for(c=1;c<256;c++) {
			map[c].r=LRAND()%65535;
			map[c].b=LRAND()%65535;
			map[c].g=LRAND()%65535; 
		}
		map[0].r=map[0].g=map[0].b=0;
		ggiSetPaletteVec(visual,0,256,map);
		ggiFillscreen(visual);
		n = 300;
	} else {
		if (halfrandom(5))
			return;
		if (halfrandom(5))
			n = 100;
		else
			n = 2000;
	}

	ggiSetGCForeground(visual, halfrandom(256));

	for (i = 0; i < n; i++) {
		if (not_solid)
			ggiSetGCForeground(visual, halfrandom(256));
		ggiDrawBox(visual,
			       halfrandom(s->width - w),
			       halfrandom(s->height - w),
			       w, w);
	}
	s->first_time = 0;
}

static int
quantize(double d)
{
	int         i = (int) floor(d);
	double      f = d - i;

	if ((LRAND() & 0xff) < f * 0xff)
		i++;
	return i;
}

void
init_slip(int xsize, int ysize)
{
	slipstruct *sp;

	if (slips == NULL) {
		if ((slips = (slipstruct *) malloc(sizeof (slipstruct))) == NULL)
			return;
	}
	sp = &slips[0];

	sp->width = xsize;
	sp->height = ysize;

	sp->blit_width = sp->width / 25;
	sp->blit_height = sp->height / 25;
	sp->nblits_remaining = 0;
	sp->mode = 0;
	sp->first_time = 1;
}

void
draw_slip(ggi_visual_t visual, int xsize, int ysize)
{
	slipstruct *s = &slips[0];
	int         timer;

	timer = 10;

	while (timer--) {
		int         xi = halfrandom(s->width - s->blit_width);
		int         yi = halfrandom(s->height - s->blit_height);
		double      x, y, dx = 0, dy = 0, t, s1, s2;

		if (0 == s->nblits_remaining--) {
			static      lut[] =
			{0, 0, 0, 1, 1, 1, 2};

			prepare_screen(visual, xsize, ysize, s);
			s->nblits_remaining = 10 *
				(2000 + halfrandom(1000) + halfrandom(1000));
			if (s->mode == 2)
				s->mode = halfrandom(2);
			else
				s->mode = lut[halfrandom(7)];
		}
		x = (2 * xi + s->blit_width) / (double) s->width - 1;
		y = (2 * yi + s->blit_height) / (double) s->height - 1;

		/* (x,y) is in biunit square */
		switch (s->mode) {
			case 0:
				dx = x;
				dy = y;

				if (dy < 0) {
					dy += 0.04;
					if (dy > 0)
						dy = 0.00;
				}
				if (dy > 0) {
					dy -= 0.04;
					if (dy < 0)
						dy = 0.00;
				}
				t = dx * dx + dy * dy + 1e-10;
				s1 = 2 * dx * dx / t - 1;
				s2 = 2 * dx * dy / t;

				dx = s1 * 5;
				dy = s2 * 5;
				break;
			case 1:
				dx = erandom(3);
				dy = erandom(3);
				break;
			case 2:
				dx = x * 3;
				dy = y * 3;
				break;
		}
		{
			int         qx = xi + quantize(dx), qy = yi + quantize(dy);
			int         wrap;

			if (qx < 0 || qy < 0 ||
			    qx >= s->width - s->blit_width ||
			    qy >= s->height - s->blit_height)
				continue;

			ggiCopyBox(visual, xi, yi,
				  s->blit_width, s->blit_height,
				  qx, qy);

			switch (s->mode) {
				case 0:
					/* wrap */
					wrap = s->width - (2 * s->blit_width);
					if (qx > wrap)
						ggiCopyBox(visual, qx, qy,
						s->blit_width, s->blit_height,
							  qx - wrap, qy);

					if (qx < 2 * s->blit_width)
						ggiCopyBox(visual, qx, qy,
						s->blit_width, s->blit_height,
							  qx + wrap, qy);

					wrap = s->height - (2 * s->blit_height);
					if (qy > wrap)
						ggiCopyBox(visual, qx, qy,
						s->blit_width, s->blit_height,
							  qx, qy - wrap);

					if (qy < 2 * s->blit_height)
						ggiCopyBox(visual, qx, qy,
						s->blit_width, s->blit_height,
							  qx, qy + wrap);
					break;
				case 1:
				case 2:
					break;
			}
		}
	}
}

void ge_slip(ggi_visual_t visual)
{
	ggi_mode mode;
	int xsize, ysize;
	ggiGetMode(visual,&mode);
	xsize=mode.visible.x;
	ysize=mode.visible.y;

	init_slip(xsize, ysize);
	while(!switchaway&&!ggiKbhit(visual)) {
		draw_slip(visual, xsize, ysize);
	}
}
